/**
 * Source: $(LINK2 https://github.com/dlang/dmd/blob/master/src/dmd/backend/md5.d, backend/_md5.d)
 */

module dmd.backend.md5;

// Online documentation: https://dlang.org/phobos/dmd_backend_md5.html

/*
 **********************************************************************
 ** md5.h -- Header file for implementation of MD5                   **
 ** RSA Data Security, Inc. MD5 Message Digest Algorithm             **
 ** Created: 2/17/90 RLR                                             **
 ** Revised: 12/27/90 SRD,AJ,BSK,JT Reference C version              **
 ** Revised (for MD5): RLR 4/27/91                                   **
 **   -- G modified to have y&~z instead of y&z                      **
 **   -- FF, GG, HH modified to add in last register done            **
 **   -- Access pattern: round 2 works mod 5, round 3 works mod 3    **
 **   -- distinct additive constant for each step                    **
 **   -- round 4 added, working mod 7                                **
 **********************************************************************
 */

/*
 **********************************************************************
 ** Copyright (C) 1990, RSA Data Security, Inc. All rights reserved. **
 **                                                                  **
 ** License to copy and use this software is granted provided that   **
 ** it is identified as the "RSA Data Security, Inc. MD5 Message     **
 ** Digest Algorithm" in all material mentioning or referencing this **
 ** software or this function.                                       **
 **                                                                  **
 ** License is also granted to make and use derivative works         **
 ** provided that such works are identified as "derived from the RSA **
 ** Data Security, Inc. MD5 Message Digest Algorithm" in all         **
 ** material mentioning or referencing the derived work.             **
 **                                                                  **
 ** RSA Data Security, Inc. makes no representations concerning      **
 ** either the merchantability of this software or the suitability   **
 ** of this software for any particular purpose.  It is provided "as **
 ** is" without express or implied warranty of any kind.             **
 **                                                                  **
 ** These notices must be retained in any copies of any part of this **
 ** documentation and/or software.                                   **
 **********************************************************************
 */

extern (C++):
nothrow:
@nogc:

/* typedef a 32 bit type */
alias UINT4 = uint;

/* Data structure for MD5 (Message Digest) computation */
struct MD5_CTX {
  UINT4[2] i;                   /* number of _bits_ handled mod 2^64 */
  UINT4[4] buf;                                    /* scratch buffer */
  ubyte[64] _in;                                     /* input buffer */
  ubyte[16] digest;             /* actual digest after MD5Final call */
}

void MD5Init (MD5_CTX *mdContext);
void MD5Update (MD5_CTX *mdContext, ubyte *inBuf, uint inLen);
void MD5Final (MD5_CTX *mdContext);

/*
 **********************************************************************
 ** End of md5.h                                                     **
 ******************************* (cut) ********************************
 */
